{*******************************************************}
{                                                       }
{                   fspLocalizer  v 1.0.1               }
{                                                       }
{             Copyright (c) 2013 FSPro Labs             }
{              http://delphi.fsprolabs.com              }
{                                                       }
{*******************************************************}


{$IFDEF CONDITIONALEXPRESSIONS}
   {$IF CompilerVersion >= 20.0}
      {$DEFINE D2009UP}
   {$IFEND}
{$ENDIF}

{$IFDEF D2009UP}
      {$DEFINE USEMEMINI}
{$ENDIF}

// Old versions of Delphi (at least since version 5) have TMemIniFile class,  but withou unicode support.
// TIniFile is based on WinAPI functions and can read unicode Ini files.
// So it is beeter to use TIniFiles with old Delphi if your application supports Unicode (e.g. with TNT).
// If your application doesn't support Unicode, you can use TMemIniFile
// instead of TIniFile - in such a case DEFINE USEMEMINI unconditionally.


unit fspLocalizer;

interface
uses Inifiles, Forms;

const
  cnstCommonMessages = 'CommonMessages';

type
  TFspLocalizer = class(TObject)
  private
   {$IFDEF USEMEMINI}
    FIniFile : TMemIniFile;
   {$ELSE}
    FIniFile : TIniFile;
   {$ENDIF}
    function GetFileName: String;
  public
    constructor Create;
    destructor Destroy; virtual;
    procedure SetLanguage(FileName : String);
    procedure SaveForm(Form: TCustomForm; Section: String = '');
    procedure LoadForm(Form: TCustomForm; Section: String);
    function GetValue(Section : String; Key: String) : String;
    procedure SetValue(Section : String; Key: String; Value : String);
    function GetCommonMessage(Key: String) : String;
    procedure SetCommonMessage(Key: String; Value : String);
    procedure Flush;
    property FileName : String read GetFileName;
  end;

var
  FsproLocalizer : TFspLocalizer;

implementation
uses SysUtils, Controls, Classes, TypInfo;

const
  SaveProps : Array[0..0 {1}] of String = ('Caption' {, 'Text'});

function GetStrPropertyValue(Obj : TObject; PropName : String) : String;
var
  PropInfo: PPropInfo;
begin
  Result := '';
  Propinfo := getPropInfo(Obj, PropName, [tkString, tkLString, tkWString {$IFDEF D2009UP}, tkUString{$ENDIF}]);
  if Propinfo = nil then
    Exit;
  Result := GetStrProp(Obj, PropInfo);
end;

///////////////////////////////////////////////////
////////////// SetPropertyValue ///////////////////
//  You can set Key to almost any published string property, e.g.
// 'Caption' - it's a form (TopObject) caption
// 'Label1.Caption'
// 'Font.Name'
// 'Button1.Font.Name'
// 'Memo1.Lines.Text'  - TStrings.Text is the only unpublished property accepted by the localizer
// 'Memo1.Lines[0]  (works only for TStrings)
// 'ListView1.Columns[0].Caption'
function SetPropertyValue(TopObject : TObject; Key : String; Value : String) : Boolean;

  function CheckIndexed(var Name : String; var IndexValue : Integer) : Boolean;
  var
    Ps, Len : Integer;
  begin
    Result := False;
    IndexValue := -1;
    Len := Length(Name);
    if Name[Len] = ']' then
    begin
      Ps := Pos('[', Name);
      if Ps <> 0 then
      begin
        try
          IndexValue := StrToInt(Copy(Name, Ps + 1, Len - Ps - 1));
          Name := Copy(Name, 1, Ps - 1);
          Result := True;
        except
          Result := False;
        end;
      end;
    end;
  END;

  function InternalSetPropertyValue(TopObj : TObject; KeyList : TStringList; Value : String) : Boolean;
  var
    NewTopObjStr : String;
    NewTopObj : TObject;
    Indexed : Boolean;
    IndexValue : Integer;
    PropName : String;
    IndexedObj : TObject;
    CollectionItem : TObject;
  begin
    Result := False;
    if KeyList.Count = 0 then
      Exit; // Something is worng. We should never execute this line

    if KeyList.Count = 1 then begin
     //end of recusion
      PropName := KeyList[0];
      Indexed := CheckIndexed(PropName, IndexValue); // Checking indexed
      if Indexed then begin
       //Setting value to indexed property is realized for TStrings only
        if GetPropInfo(TopObj, PropName, [tkClass]) = nil then
          Exit; // Must be a class
        IndexedObj := GetObjectProp(TopObj, PropName);
        if IndexedObj is TStrings then
          try
            TStrings(IndexedObj)[IndexValue] := Value;
            Result := True;
          except
          end;
      end
      else begin
        if (TopObj is TStrings) and SameText(PropName, 'Text') then
          TStrings(TopObj).Text := Value
        else
          try
            SetStrProp(TopObj, PropName, Value); // or SetPropValue for all types.
            Result := True;
          except
          end;
      end;
    end
    else begin
      NewTopObjStr := KeyList[0];
      KeyList.Delete(0);
      Indexed := CheckIndexed(NewTopObjStr, IndexValue);
      NewTopObj := nil;
      if GetPropInfo(TopObj, NewTopObjStr, [tkClass]) <> nil  then begin
        NewTopObj := GetObjectProp(TopObj, NewTopObjStr);
        if Indexed then
          try
            NewTopObj := TCollection(NewTopObj).Items[IndexValue];
          except
            NewTopObj := nil;
          end;
      end
      else
        if TopObj is TComponent then
          NewTopObj := TComponent(TopObj).FindComponent(NewTopObjStr);
      if NewTopObj = nil then
        Exit; //Object not found
      InternalSetPropertyValue(NewTopObj, KeyList, Value);
    end;
  end;

var
  Elements : TStringList;
  NewKey : String;
  i      : Integer;
  Control : TComponent;
  Prop    : String;
  ControlStr : String;

begin
  Elements := TStringList.Create;
  try
    NewKey := StringReplace(Key, '.', ',', [rfReplaceAll]);
    NewKey := StringReplace(NewKey, ':', ',', [rfReplaceAll]);
    Elements.CommaText := NewKey;
    Result := InternalSetPropertyValue(TopObject, Elements, Value);
  finally
    Elements.Free;
  end;
end;


{ TFspLocalizer }

constructor TFspLocalizer.Create;
begin
  inherited;
  FIniFile := nil;
end;

destructor TFspLocalizer.Destroy;
begin
  FIniFile.Free;
  inherited;
end;

procedure TFspLocalizer.SetLanguage(FileName: String);
begin
  if Assigned(FIniFile) then
    FIniFile.Free;
 {$IFDEF USEMEMINI}
  FIniFile := TMemIniFile.Create(FileName {$IFDEF D2009UP}, TEncoding.Unicode {$ENDIF});
 {$ELSE}
  FIniFile := TIniFile.Create(FileName);
 {$ENDIF}
end;


function TFspLocalizer.GetFileName: String;
begin
  Result := FIniFile.FileName;
end;


function TFspLocalizer.GetValue(Section, Key: String): String;
begin
  Result := FIniFile.ReadString(Section, Key, '');
end;

procedure TFspLocalizer.SetValue(Section, Key, Value: String);
begin
  FIniFile.WriteString(Section, Key, Value);
end;

function TFspLocalizer.GetCommonMessage(Key: String): String;
begin
  Result := GetValue(cnstCommonMessages, Key);
end;

procedure TFspLocalizer.SetCommonMessage(Key, Value: String);
begin
  SetValue(cnstCommonMessages, Key, Value);
end;

procedure TFspLocalizer.Flush;
begin
 {$IFDEF USEMEMINI}
  FIniFile.UpdateFile;
 {$ENDIF}
end;

procedure TFspLocalizer.SaveForm(Form : TCustomForm; Section : String);
var
  i : Integer;
  j : Integer;
  Control : TComponent;
  PropValue : String;
begin
  if FIniFile = nil then
    Exit;
  if Section = '' then
    Section := Form.ClassName;
  SetValue(Section, 'Caption',Form.Caption);

  for i := 0 to Form.ComponentCount - 1 do begin
    Control := Form.Components[i];
    for j := 0 to High(SaveProps) do begin
      PropValue := GetStrPropertyValue(Control, SaveProps[j]);
      if PropValue <> '' then
        SetValue(Section, Control.Name + '.' +  SaveProps[j],PropValue);
    end;
  end;
  Flush;
end;

procedure TFspLocalizer.LoadForm(Form : TCustomForm; Section : String);
var
  i : Integer;
  Keys : TStrings;
  Key  : String;
  Value : String;
begin
  if FIniFile = nil then
    Exit;
  if Section = '' then
    Section := Form.ClassName;
  Keys := TStringList.Create;
  try
    FIniFile.ReadSection(Section, Keys);
    for i := 0 to Keys.Count - 1 do begin
      Key := Keys[i];
      Value := GetValue(Section, Key);
      Value := StringReplace(Value, '\n', #13, [rfReplaceAll]);
      SetPropertyValue(Form, Key, Value);
    end;
  finally
    Keys.Free;
  end;
end;

initialization
  FsproLocalizer := TFspLocalizer.Create;

finalization
  FsproLocalizer.Free;

end.
